export module Jsonapi { 
export interface IAttributes {
    // problem with lines like
    /*
    this._receipt_number = order.attributes.receipt_number;
    this.transaction.attributes._due_date_with_format = new Date(this.transaction.attributes.due_date);
    */
    // [value: string]: boolean | string | number;
    [value: string]: any;
}


export interface ICache {
    setResource(resource: IResource): void;
    deprecateCollections(path_start_with: string): boolean;
}


export interface ICacheMemory extends ICache {
    resources: { [id: string]: IResource };

    getOrCreateCollection(url: string): ICollection;
    isCollectionExist(url: string): boolean;
    isCollectionLive(url: string, ttl: number): boolean;

    isResourceLive(id: string, ttl: number): boolean;
    getOrCreateResource(type: string, id: string): IResource;
    setCollection(url: string, collection: ICollection): void;

    removeResource(id: string): void;
}


export interface ICacheStore extends ICache {
    getResource(resource: IResource): ng.IPromise<object>;
    getCollectionFromStorePromise(url:string, includes: Array<string>, collection: ICollection): ng.IPromise<ICollection>;
    setCollection(url: string, collection: ICollection, include: Array<string>): void;
}


export interface ICollection extends Array<IResource> {
    $length: number;
    $toArray: Array<IResource>;
    $is_loading: boolean;
    $source: 'new' | 'memory' | 'store' | 'server';
    $cache_last_update: number;
    data: Array<IDataResource>; // this need disapear is for datacollection
    page: IPage;
}


export interface ICore {
    // jsonapiServices: Object;

    loadingsCounter: number;
    loadingsStart: Function;
    loadingsDone: Function;
    loadingsError: Function;
    loadingsOffline: Function;

    _register(clase: IService): boolean;
    getResourceService(type: string): IService;
    refreshLoadings(factor: number): void;
    clearCache(): void;
    duplicateResource(resource: IResource, ...relations_types: Array<string>): IResource;

    // static
    me?: IService;

    // defined on core code too
    injectedServices?: {
        $q: ng.IQService,
        JsonapiStoreService: any,
        JsonapiHttp: any,
        rsJsonapiConfig: any
    };
}


interface IDataCollection extends IDocument {
    data: IDataResource[];
    page?: IPage;
    _lastupdate_time?: number;   // used when come from Store
}


interface IDataObject extends IDocument {
    data: IDataResource;
}


interface IDataResource {
    type: string;
    id: string;
    attributes?: IAttributes;
    relationships?: object;
    links?: ILinks;
    meta?: object;
}


// http://org/format/#document-top-level
interface IDocument {
    // data in child interface IJsonapiCollection
    // error in child interface IJsonapiErrors
    jsonapi?: string;
    links?: ILinks;
    included?: Array<IDataResource>;
    meta?: Object;
}


interface IErrors extends IDocument {
    errors: [
        {
            code?: string,
            source?: {
                attributes?: string,
                pointer: string
            },
            title?: string,
            detail?: string
        }
    ];
}


export interface IExecParamsProcessed {
    id: string;
    params: IParamsCollection | IParamsResource;
    fc_success: Function;
    fc_error: Function;
    exec_type: 'all' | 'get' | 'delete' | 'save';
}


export interface IExecParams {
    id: string;
    params?: IParamsCollection | IParamsResource | Function;
    fc_success?: Function;
    fc_error?: Function;
    exec_type: 'all' | 'get' | 'delete' | 'save';
}


// http://org/format/#document-links
export interface ILinks {
    self?: string;
    related?: {
        href: string;
        meta: object;
    };
}

export interface IPage {
    number: number;

    // http://jsonapi.org/format/#fetching-pagination
    limit?: number;

    // multinexo
    total_resources?: number;
    resources_per_page?: number;
}


interface IParamsCollection extends IParams {
    localfilter?: object;
    remotefilter?: object;
    smartfilter?: object;
    page?: IPage;
    storage_ttl?: number;
    cachehash?: string;  // solution for when we have different resources with a same id
}


interface IParamsResource extends IParams {
    id?: string;
}

export interface IParams {
    beforepath?: string;
    include?: Array<string>;
    ttl?: number;
}


interface IRelationship {
    // IDataResource added for this reason:
    // redefined from IDataResource (return errors /home/rsk/desarrollo/ts-angular-jsonapi/src/library/services/resource-functions.ts)
    data: ICollection | IResource | IDataResource | {};
}


interface IRelationships {
    [value: string]: IRelationship;
}


export interface IResource {
    type: string;
    id: string;
    attributes?: IAttributes;
    relationships: IRelationships;    // redefined from IDataResource

    is_new: boolean;
    is_loading: boolean;
    is_saving: boolean;
    lastupdate?: number;


    reset? (): void;
    addRelationship? (resource: IResource, type_alias?: string): void;
    addRelationships? (resources: ICollection, type_alias: string): void;
    removeRelationship? (type_alias: string, id: string): boolean;
    addRelationshipsArray <T extends IResource>(resources: Array<T>, type_alias?: string): void;
    save<T extends IResource>(params?: IParamsResource, fc_success?: Function, fc_error?: Function): ng.IPromise<object>;
    toObject? (params?: IParamsResource): IDataObject;
    getService(): IService;
}


export interface IResourcesById {
    [resource_id: string]: IResource;
}


export interface IResourcesByType {
    [type: string]: IResourcesById;
}

export interface ISchema {
    attributes?: object;
    relationships?: object;
    ttl?: number;
    path?: string;
}


export interface IService {
    type: string;
    schema: ISchema;
    getPrePath(): string;
    getPath(): string;
    register(): boolean;
    get<T extends IResource>(id: string | number, params?: IParamsResource | Function, fc_success?: Function, fc_error?: Function): T;
    all(params?: IParamsCollection | Function, success?: Function, error?: Function): ICollection;
    delete (id: String, params?: IParamsResource | Function, success?: Function, error?: Function): void;
    getService<T extends IService> ():T;
    clearCacheMemory? (): boolean;
    new?<T extends IResource>(): T;
    cachememory: ICacheMemory;
    cachestore: ICacheStore;
    parseFromServer(attributes: IAttributes): void;
    parseToServer?(attributes: IAttributes): void;
}

export class Core implements ICore {
    protected rsJsonapiConfig: any;
    protected JsonapiCoreServices: any;
    static me: ICore;
    static injectedServices: {
        $q: ng.IQService;
        JsonapiStoreService: any;
        JsonapiHttp: any;
        rsJsonapiConfig: any;
    };
    private resourceServices;
    loadingsCounter: number;
    loadingsStart: () => void;
    loadingsDone: () => void;
    loadingsError: () => void;
    loadingsOffline: () => void;
    /** @ngInject */
    constructor(rsJsonapiConfig: any, JsonapiCoreServices: any);
    _register(clase: IService): boolean;
    getResourceService(type: string): IService;
    refreshLoadings(factor: number): void;
    clearCache(): boolean;
    duplicateResource(resource: IResource, ...relations_alias_to_duplicate_too: Array<string>): IResource;
}


export class ParentResourceService {
    protected proccess_exec_params(exec_params: IExecParams): IExecParamsProcessed;
    protected runFc(some_fc: any, param: any): any;
}

export class Resource extends ParentResourceService implements IResource {
    is_new: boolean;
    is_loading: boolean;
    is_saving: boolean;
    id: string;
    type: string;
    attributes: IAttributes;
    relationships: IRelationships;
    reset(): void;
    toObject(params?: IParamsResource): IDataObject;
    save<T extends IResource>(params?: Object | Function, fc_success?: Function, fc_error?: Function): ng.IPromise<object>;
    protected __exec<T extends IResource>(exec_params: IExecParams): ng.IPromise<object>;
    private _save<T>(params, fc_success, fc_error);
    addRelationship<T extends IResource>(resource: T, type_alias?: string): void;
    addRelationships(resources: ICollection, type_alias: string): void;
    addRelationshipsArray<T extends IResource>(resources: Array<T>, type_alias?: string): void;
    removeRelationship(type_alias: string, id: string): boolean;
    getService(): IService;
}

export class Service extends ParentResourceService implements IService {
    schema: ISchema;
    cachememory: ICacheMemory;
    cachestore: ICacheStore;
    type: string;
    private path;
    private smartfiltertype;
    register(): boolean;
    newResource(): IResource;
    new<T extends IResource>(): T;
    getPrePath(): string;
    getPath(): string;
    get<T extends IResource>(id: any, params?: IParamsResource | Function, fc_success?: Function, fc_error?: Function): T;
    delete(id: string, params?: Object | Function, fc_success?: Function, fc_error?: Function): void;
    all(params?: IParamsCollection | Function, fc_success?: Function, fc_error?: Function): ICollection;
    protected __exec(exec_params: IExecParams): IResource | ICollection | void;
    _get(id: string, params: IParamsResource, fc_success: any, fc_error: any): IResource;
    private getGetFromServer(path, fc_success, fc_error, resource);
    private _all(params, fc_success, fc_error);
    private getAllFromServer(path, params, fc_success, fc_error, tempororay_collection, cached_collection);
    private _delete(id, params, fc_success, fc_error);
    getService<T extends IService>(): T;
    clearCacheMemory(): boolean;
    parseFromServer(attributes: IAttributes): void;
}
}